"""
# 3D Cellpose Extension.
# Copyright (C) 2021 D. Eschweiler, J. Stegmaier
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the Liceense at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Please refer to the documentation for more information about the software
# as well as for installation instructions.
"""

import os
import json
import torch
import numpy as np


# Creates a default augmentation dict at the specified location
def create_defaults(save_dir):

    augmentation_dict = {}
    augmentation_dict["prob"] = 0.5
    augmentation_dict["queue"] = [
        "scale",
        "noise",
        "shuffle",
        "inpaint",
        "decline",
        "variance_sampling",
        "stretch",
        "shrink",
        "anisotropy",
    ]

    augmentation_dict["scale_min"] = 0.6
    augmentation_dict["scale_max"] = 1.2

    augmentation_dict["noise_mean"] = 0
    augmentation_dict["noise_std"] = 0.2

    augmentation_dict["shuffle_size"] = (25, 25, 25)
    augmentation_dict["shuffle_count"] = 3

    augmentation_dict["inpaint_size"] = (15, 15, 15)
    augmentation_dict["inpaint_count"] = 3

    augmentation_dict["decline_axes"] = [0, 1, 2]

    augmentation_dict["sampling_mean_weight"] = 1
    augmentation_dict["sampling_std_weight"] = 1

    augmentation_dict["stretch_factor"] = 2
    augmentation_dict["stretch_axes"] = [0, 1, 2]

    augmentation_dict["shrink_factor"] = 2
    augmentation_dict["shrink_axes"] = [0, 1, 2]

    augmentation_dict["anisotropy_factor"] = 2
    augmentation_dict["anisotropy_axes"] = [0, 1, 2]

    augmentation_dict["permute_dim"] = True

    with open(os.path.join(save_dir, "augments_default.json"), "w") as file_handle:
        json.dump(augmentation_dict, file_handle)


# Intensity augmentations
# Usually only used for input images
class intensity_augmenter:
    def __init__(self, augmentation_dict={}):

        self.augmentation_dict = augmentation_dict
        self.dict_sanitycheck()

    def dict_sanitycheck(self):

        # main parameters
        _ = self.augmentation_dict.setdefault("prob", 0)
        _ = self.augmentation_dict.setdefault("queue", [])

        # linear scaling
        _ = self.augmentation_dict.setdefault("scale_min", 0.6)
        _ = self.augmentation_dict.setdefault("scale_max", 1.2)

        # additive noise
        _ = self.augmentation_dict.setdefault("noise_mean", 0)
        _ = self.augmentation_dict.setdefault("noise_std", 0.1)

        # windowed shuffle
        _ = self.augmentation_dict.setdefault("shuffle_size", (25, 25, 25))
        _ = self.augmentation_dict.setdefault("shuffle_count", 3)
        assert (
            len(self.augmentation_dict["shuffle_size"]) == 3
        ), "Shuffle window size must be 3-dimensional."

        # paint-in
        _ = self.augmentation_dict.setdefault("inpaint_size", (15, 15, 15))
        _ = self.augmentation_dict.setdefault("inpaint_count", 3)
        assert (
            len(self.augmentation_dict["inpaint_size"]) == 3
        ), "Inpainting window size must be 3-dimensional."

        # decline
        _ = self.augmentation_dict.setdefault("decline_axes", [0, 1, 2])
        if not isinstance(self.augmentation_dict["decline_axes"], (list, tuple)):
            self.augmentation_dict["decline_axes"] = list(
                self.augmentation_dict["decline_axes"]
            )

    # Linear scaling drawn from a uniform distribution
    def linear_scaling(self, patch, min_val=1, max_val=1):
        patch_min, patch_max = np.min(patch), np.max(patch)
        patch = patch * np.random.uniform(low=min_val, high=max_val)
        patch = np.clip(patch, patch_min, patch_max)
        return patch

    # Additive Gaussian noise
    def additive_noise(self, patch, mean=0, std=0):
        patch_min, patch_max = np.min(patch), np.max(patch)
        patch = patch + np.random.normal(loc=mean, scale=std, size=patch.shape)
        patch = np.clip(patch, patch_min, patch_max)
        return patch

    # Windowed shuffle
    def windowed_shuffle(self, patch, shuffle_size=(5, 5, 5), shuffle_count=0):

        for num_window in range(shuffle_count):

            # get the current window coordinates
            window_start = [
                np.random.randint(0, np.maximum(1, patch_dim - window_dim))
                for window_dim, patch_dim in zip(shuffle_size, patch.shape)
            ]
            window_end = [
                start + window_dim
                for start, window_dim in zip(window_start, shuffle_size)
            ]
            slicing = tuple(map(slice, window_start, window_end))

            # crop the window and shuffle its content
            window = patch[slicing]
            np.random.shuffle(window)

            # replace the current values with the shuffled ones
            patch[slicing] = window

        return patch

    # Inpainting
    def inpaint(self, patch, inpaint_size=(5, 5, 5), inpaint_count=0):
        assert len(inpaint_size) == 3, "Window size must be 3-dimensional."
        for num_window in range(inpaint_count):

            # get the current window coordinates
            window_start = [
                np.random.randint(0, np.maximum(1, patch_dim - window_dim))
                for window_dim, patch_dim in zip(inpaint_size, patch.shape)
            ]
            window_end = [
                start + window_dim
                for start, window_dim in zip(window_start, inpaint_size)
            ]
            slicing = tuple(map(slice, window_start, window_end))

            # print in the current window location
            patch[slicing] = np.random.uniform(low=patch.min(), high=patch.max())

        return patch

    # Intensity decline
    def intensity_decline(self, patch, decline_axes=[0]):

        # get a random dimension
        decline_axis = np.random.choice(decline_axes)

        # define the decline array
        decline_extend = np.random.uniform(0, 1)
        decline = np.linspace(0, 1, num=int(decline_extend * patch.shape[decline_axis]))
        decline = np.pad(
            decline, (0, patch.shape[decline_axis] - len(decline)), constant_values=1
        )
        decline = np.expand_dims(
            decline, axis=tuple([i for i in range(patch.ndim) if not i == decline_axis])
        )

        # apply the decline transformation
        patch = patch * decline

        return patch

    def apply(self, patch):

        if np.any(np.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the scale augmentation.'
            )
        if (
            np.random.rand() <= self.augmentation_dict["prob"]
            and "scale" in self.augmentation_dict["queue"]
        ):
            patch = self.linear_scaling(
                patch,
                min_val=self.augmentation_dict["scale_min"],
                max_val=self.augmentation_dict["scale_max"],
            )

        if np.any(np.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the noise augmentation.'
            )
        if (
            np.random.rand() <= self.augmentation_dict["prob"]
            and "noise" in self.augmentation_dict["queue"]
        ):
            patch = self.additive_noise(
                patch,
                mean=self.augmentation_dict["noise_mean"],
                std=self.augmentation_dict["noise_std"],
            )

        if np.any(np.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the shuffle augmentation.'
            )
        if (
            np.random.rand() <= self.augmentation_dict["prob"]
            and "shuffle" in self.augmentation_dict["queue"]
        ):
            patch = self.windowed_shuffle(
                patch,
                shuffle_size=self.augmentation_dict["shuffle_size"],
                shuffle_count=self.augmentation_dict["shuffle_count"],
            )

        if np.any(np.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the inpaint augmentation.'
            )
        if (
            np.random.rand() <= self.augmentation_dict["prob"]
            and "inpaint" in self.augmentation_dict["queue"]
        ):
            patch = self.inpaint(
                patch,
                inpaint_size=self.augmentation_dict["inpaint_size"],
                inpaint_count=self.augmentation_dict["inpaint_count"],
            )

        if np.any(np.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the decline augmentation.'
            )
        if (
            np.random.rand() <= self.augmentation_dict["prob"]
            and "decline" in self.augmentation_dict["queue"]
        ):
            patch = self.intensity_decline(
                patch, decline_axes=self.augmentation_dict["decline_axes"]
            )

        return patch


class geometry_augmenter:
    def __init__(self, augmentation_dict={}):

        self.augmentation_dict = augmentation_dict
        self.dict_sanitycheck()

        # Preset random probabilities
        self.prob_anisotropy = np.random.rand()
        self.prob_stretching = np.random.rand()

    def dict_sanitycheck(self):

        # main parameters
        _ = self.augmentation_dict.setdefault("prob", 0)
        _ = self.augmentation_dict.setdefault("queue", [])

        # stretching
        _ = self.augmentation_dict.setdefault("stretch_factor", 2)
        _ = self.augmentation_dict.setdefault("stretch_axes", [0, 1, 2])
        if not isinstance(self.augmentation_dict["stretch_axes"], (list, tuple)):
            self.augmentation_dict["stretch_axes"] = list(
                self.augmentation_dict["stretch_axes"]
            )
        _ = self.augmentation_dict.setdefault(
            "stretch_axis", self.augmentation_dict["stretch_axes"][0]
        )

        # shrinking
        _ = self.augmentation_dict.setdefault("shrink_factor", 2)
        _ = self.augmentation_dict.setdefault("shrink_axes", [0, 1, 2])
        if not isinstance(self.augmentation_dict["shrink_axes"], (list, tuple)):
            self.augmentation_dict["shrink_axes"] = list(
                self.augmentation_dict["shrink_axes"]
            )
        _ = self.augmentation_dict.setdefault(
            "shrink_axis", self.augmentation_dict["shrink_axes"][0]
        )

        # anisotropy
        _ = self.augmentation_dict.setdefault("anisotropy_factor", 2)
        _ = self.augmentation_dict.setdefault("anisotropy_axes", [0, 1, 2])
        if self.augmentation_dict["anisotropy_factor"] % 1 > 0:
            print(
                'The anisotropy factor needs to be integer valued. Setting "anisotropy_factor" to {}.'.format(
                    int(self.augmentation_dict["anisotropy_factor"])
                )
            )
        self.augmentation_dict["anisotropy_factor"] = int(
            self.augmentation_dict["anisotropy_factor"]
        )
        if not isinstance(self.augmentation_dict["anisotropy_axes"], (list, tuple)):
            self.augmentation_dict["anisotropy_axes"] = list(
                self.augmentation_dict["anisotropy_axes"]
            )
        _ = self.augmentation_dict.setdefault(
            "anisotropy_axis", self.augmentation_dict["anisotropy_axes"][0]
        )

        # random dimension permutation (needs to be implemented in the data loader)
        _ = self.augmentation_dict.setdefault("permute_dim", False)

    # Stretching
    def stretching(self, patch, stretch_factor=1, stretch_axis=0):

        # save the current shape
        orig_shape = patch.shape

        # stretch a cropped portion of the patch
        patch_crop = [
            int(np.ceil(p / stretch_factor)) if stretch_axis == i else p
            for i, p in enumerate(patch.shape)
        ]
        patch = patch[: patch_crop[0], : patch_crop[1], : patch_crop[2]]
        patch = np.repeat(patch, stretch_factor, axis=stretch_axis)
        patch = patch[: orig_shape[0], : orig_shape[1], : orig_shape[2]]

        # Sanity check
        assert (
            patch.shape == orig_shape
        ), f"Shape missmatch after stretching. {orig_shape} to {patch.shape}."

        return patch

    # Shrinking
    def shrinking(self, patch, shrink_factor=1, shrink_axis=0):

        # sabe the current shape
        orig_shape = patch.shape

        # shrink the whole image with the given factor
        # use reflected padding to reconstruct the whole patch size
        if shrink_axis == 0:
            patch = patch[::shrink_factor, ...]
        elif shrink_axis == 1:
            patch = patch[:, ::shrink_factor, ...]
        elif shrink_axis == 2:
            patch = patch[:, :, ::shrink_factor, ...]
        else:
            raise NotImplementedError(
                f'Anisotropy transform not implemented for axis "{shrink_axis}".'
            )

        # pad the patch to the original size
        padding = [(0, o - p) for o, p in zip(orig_shape, patch.shape)]
        patch = np.pad(patch, padding, mode="reflect")

        # Sanity check
        assert (
            patch.shape == orig_shape
        ), f"Shape missmatch after shrinking transform. {orig_shape} to {patch.shape}."

        return patch

    # Anisotropy
    def anisotropy(self, patch, anisotropy_factor=1, anisotropy_axis=0):

        # save the current shape
        orig_shape = patch.shape

        # extraxt every n-th slice
        if anisotropy_axis == 0:
            patch = patch[::anisotropy_factor, ...]
        elif anisotropy_axis == 1:
            patch = patch[:, ::anisotropy_factor, ...]
        elif anisotropy_axis == 2:
            patch = patch[:, :, ::anisotropy_factor, ...]
        else:
            raise NotImplementedError(
                f'Anisotropy transform not implemented for axis "{anisotropy_axis}".'
            )

        # repeat each slice n-times
        patch = np.repeat(patch, anisotropy_factor, axis=anisotropy_axis)

        # Sanity check
        assert (
            patch.shape == orig_shape
        ), f"Shape missmatch after anisotropy transform. {orig_shape} to {patch.shape}."

        return patch

    def apply(self, patch, reset=True):

        # `reset=True` chooses new values for the random components of each transformation
        # Otherwise, the same transformations are used again (e.g., same transformations for image/mask pairs)

        if np.any(np.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the anisotropy augmentation.'
            )
        if reset:
            self.prob_anisotropy = np.random.rand()
            self.augmentation_dict["anisotropy_axis"] = np.random.choice(
                self.augmentation_dict["anisotropy_axes"]
            )
        if (
            self.prob_anisotropy <= self.augmentation_dict["prob"]
            and "anisotropy" in self.augmentation_dict["queue"]
        ):
            patch = self.anisotropy(
                patch,
                anisotropy_factor=self.augmentation_dict["anisotropy_factor"],
                anisotropy_axis=self.augmentation_dict["anisotropy_axis"],
            )

        if np.any(np.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the stretching augmentation.'
            )
        if reset:
            self.prob_stretching = np.random.rand()
            self.augmentation_dict["stretch_axis"] = np.random.choice(
                self.augmentation_dict["stretch_axes"]
            )
        if (
            self.prob_stretching <= self.augmentation_dict["prob"]
            and "stretch" in self.augmentation_dict["queue"]
        ):
            patch = self.stretching(
                patch,
                stretch_factor=self.augmentation_dict["stretch_factor"],
                stretch_axis=self.augmentation_dict["stretch_axis"],
            )

        if np.any(np.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the shrinking augmentation.'
            )
        if reset:
            self.prob_shrinking = np.random.rand()
            self.augmentation_dict["shrink_axis"] = np.random.choice(
                self.augmentation_dict["shrink_axes"]
            )
        if (
            self.prob_shrinking <= self.augmentation_dict["prob"]
            and "shrink" in self.augmentation_dict["queue"]
        ):
            patch = self.shrinking(
                patch,
                shrink_factor=self.augmentation_dict["shrink_factor"],
                shrink_axis=self.augmentation_dict["shrink_axis"],
            )

        return patch


class intensity_augmenter_pytorch:
    def __init__(self, augmentation_dict={}):

        self.augmentation_dict = augmentation_dict
        self.dict_sanitycheck()

    def dict_sanitycheck(self):

        # main parameters
        _ = self.augmentation_dict.setdefault("prob", 0)
        _ = self.augmentation_dict.setdefault("queue", [])

        # linear scaling
        _ = self.augmentation_dict.setdefault("scale_min", 0.6)
        _ = self.augmentation_dict.setdefault("scale_max", 1.2)

        # additive noise
        _ = self.augmentation_dict.setdefault("noise_mean", 0)
        _ = self.augmentation_dict.setdefault("noise_std", 0.2)

        # windowed shuffle
        _ = self.augmentation_dict.setdefault("shuffle_size", (25, 25, 25))
        _ = self.augmentation_dict.setdefault("shuffle_count", 3)
        assert (
            len(self.augmentation_dict["shuffle_size"]) == 3
        ), "Shuffle window size must be 3-dimensional."

        # decline
        _ = self.augmentation_dict.setdefault("decline_axes", [0, 1, 2])
        if not isinstance(self.augmentation_dict["decline_axes"], (list, tuple)):
            self.augmentation_dict["decline_axes"] = list(
                self.augmentation_dict["decline_axes"]
            )

    # Linear scaling drawn from a uniform distribution
    def linear_scaling(self, patch, min_val=1, max_val=1):
        patch = patch * (
            torch.tensor(min_val, dtype=patch.dtype, device=patch.device)
            + max_val * torch.rand(1, dtype=patch.dtype, device=patch.device)
        )
        return patch

    # Additive Gaussian noise
    def additive_noise(self, patch, mean=0, std=0):
        patch = (
            patch
            + torch.randn_like(patch) * (std**0.5)
            + torch.tensor(mean, dtype=patch.dtype, device=patch.device)
        )
        return patch

    # Windowed intensity scaling
    def windowed_scaling(self, patch, shuffle_size=(5, 5, 5), shuffle_count=0):

        for num_window in range(shuffle_count):

            # get the current window coordinates
            window_start = [0,] * (patch.ndim - 3) + [
                np.random.randint(0, np.maximum(1, patch_dim - window_dim))
                for window_dim, patch_dim in zip(
                    shuffle_size, patch.shape[patch.ndim - 3 :]
                )
            ]
            window_end = list(patch.shape[0 : patch.ndim - 3]) + [
                start + window_dim
                for start, window_dim in zip(
                    window_start[patch.ndim - 3 :], shuffle_size
                )
            ]
            slicing = tuple(map(slice, window_start, window_end))

            # scale the current window values
            window = torch.ones_like(patch)
            window[slicing] = torch.rand(
                tuple(np.array(window_end) - np.array(window_start)),
                dtype=patch.dtype,
                device=patch.device,
            )
            patch = patch * window

        return patch

    # Intensity decline
    def intensity_decline(self, patch, decline_axes=[0]):

        # get a random dimension
        decline_axis = decline_axes[torch.randint(0, len(decline_axes), (1,))]

        # define the decline array
        decline_extend = torch.rand(1, dtype=patch.dtype, device=patch.device)
        decline = torch.linspace(
            0,
            1,
            steps=int(decline_extend * patch.shape[decline_axis]),
            dtype=patch.dtype,
            device=patch.device,
        )
        decline = torch.nn.functional.pad(
            decline,
            (0, patch.shape[decline_axis] - len(decline)),
            mode="constant",
            value=1,
        )
        for expand_dim in [i for i in range(patch.ndim) if not i == decline_axis]:
            decline = torch.unsqueeze(decline, expand_dim)

        # apply the decline transformation
        patch = patch * decline

        return patch

    def apply(self, patch):

        if torch.any(torch.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the scale augmentation.'
            )
        if (
            torch.rand(1) <= self.augmentation_dict["prob"]
            and "scale" in self.augmentation_dict["queue"]
        ):
            patch = self.linear_scaling(
                patch,
                min_val=self.augmentation_dict["scale_min"],
                max_val=self.augmentation_dict["scale_max"],
            )

        if torch.any(torch.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the noise augmentation.'
            )
        if (
            torch.rand(1) <= self.augmentation_dict["prob"]
            and "noise" in self.augmentation_dict["queue"]
        ):
            patch = self.additive_noise(
                patch,
                mean=self.augmentation_dict["noise_mean"],
                std=self.augmentation_dict["noise_std"],
            )

        if torch.any(torch.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the shuffle augmentation.'
            )
        if (
            torch.rand(1) <= self.augmentation_dict["prob"]
            and "shuffle" in self.augmentation_dict["queue"]
        ):
            patch = self.windowed_scaling(
                patch,
                shuffle_size=self.augmentation_dict["shuffle_size"],
                shuffle_count=self.augmentation_dict["shuffle_count"],
            )

        if torch.any(torch.isnan(patch)):
            raise ValueError(
                'Encountered "NaN" value before applying the decline augmentation.'
            )
        if (
            torch.rand(1) <= self.augmentation_dict["prob"]
            and "decline" in self.augmentation_dict["queue"]
        ):
            patch = self.intensity_decline(
                patch, decline_axes=self.augmentation_dict["decline_axes"]
            )

        return patch



import os
import numpy as np
from pathlib import Path
from tifffile import imread, imwrite
from skimage.measure import regionprops
from skimage.morphology import remove_small_objects
from scipy.ndimage import binary_erosion
from tqdm import tqdm
from datetime import datetime
from uuid import uuid4


class SimplePatches:
    def __init__(
        self,
        base_dir,
        real_mask_dir,
        real_mask_patch_dir,
        patch_size,
        pattern=".tif",
        lower_ratio_fore_to_back=0.5,
        upper_ratio_fore_to_back=0.9,
    ):

       
        self.base_dir = base_dir
        self.real_mask_dir = os.path.join(base_dir, real_mask_dir)
        self.real_mask_patch_dir = os.path.join(
            self.base_dir, real_mask_patch_dir
        )
        self.patch_size = patch_size
        self.pattern = pattern
        self.lower_ratio_fore_to_back = lower_ratio_fore_to_back
        self.upper_ratio_fore_to_back = upper_ratio_fore_to_back
        self.acceptable_formats = [".tif", ".TIFF", ".TIF", ".png"]

        self._create_smart_patches()

   

    def _create_smart_patches(self):

        Path(self.real_mask_patch_dir).mkdir(exist_ok=True)
        files = os.listdir(self.real_mask_dir)
        for fname in files:
            if any(fname.endswith(f) for f in self.acceptable_formats):
                    label_image_membrane = imread(
                        os.path.join(
                            self.real_mask_dir, fname
                        )
                    ).astype(np.uint16)
                    self.ndim = len(label_image_membrane.shape)
                    properties_membrane = regionprops(label_image_membrane)
                    for count, prop in tqdm(enumerate(properties_membrane)):
                        self._label_maker(
                            fname,
                            label_image_membrane,
                            count,
                            prop,
                            self.real_mask_patch_dir,
                        )

    

    def _label_maker(
        self,
        fname: str,
        labelimage: np.ndarray,
        count: int,
        prop: regionprops,
        real_mask_patch_dir: str,
    ):

        name = os.path.splitext(fname)[0]

        if self.ndim == 2:

            self.valid = False
            centroid = prop.centroid
            x = centroid[1]
            y = centroid[0]

            crop_Xminus = x - int(self.patch_size[1] / 2)
            crop_Xplus = x + int(self.patch_size[1] / 2)
            crop_Yminus = y - int(self.patch_size[0] / 2)
            crop_Yplus = y + int(self.patch_size[0] / 2)
            crop_minus = [crop_Yminus, crop_Xminus]
            region = (
                slice(int(crop_Yminus), int(crop_Yplus)),
                slice(int(crop_Xminus), int(crop_Xplus)),
            )
        if self.ndim == 3:

            self.valid = False
            centroid = prop.centroid
            z = centroid[0]
            x = centroid[2]
            y = centroid[1]

            crop_Xminus = x - int(self.patch_size[2] / 2)
            crop_Xplus = x + int(self.patch_size[2] / 2)
            crop_Yminus = y - int(self.patch_size[1] / 2)
            crop_Yplus = y + int(self.patch_size[1] / 2)
            crop_Zminus = z - int(self.patch_size[0] / 2)
            crop_Zplus = z + int(self.patch_size[0] / 2)
            crop_minus = [crop_Zminus, crop_Yminus, crop_Xminus]
            region = (
                slice(int(crop_Zminus), int(crop_Zplus)),
                slice(int(crop_Yminus), int(crop_Yplus)),
                slice(int(crop_Xminus), int(crop_Xplus)),
            )
        if all(crop for crop in crop_minus) > 0:
            self.crop_labelimage = labelimage[region]
            self.crop_labelimage = remove_small_objects(
                self.crop_labelimage.astype("uint16"), min_size=10
            )
            if (
                self.crop_labelimage.shape[0] == self.patch_size[0]
                and self.crop_labelimage.shape[1] == self.patch_size[1]
                and self.ndim == 2
            ):
                self._crop_maker(
                    name,
                    count,
                    real_mask_patch_dir,
                )
            if (
                self.crop_labelimage.shape[0] == self.patch_size[0]
                and self.crop_labelimage.shape[1] == self.patch_size[1]
                and self.crop_labelimage.shape[2] == self.patch_size[2]
                and self.ndim == 3
            ):
                self._crop_maker(
                    name,
                    count,
                    real_mask_patch_dir,
                )

    def _crop_maker(
        self,
        name,
        count,
        real_mask_patch_dir,
    ):

        self._region_selector()
        if (
            self.valid
            
        ):
           
            
            self.eroded_crop_labelimage = self.crop_labelimage
            eventid = datetime.now().strftime("%Y%m-%d%H-%M%S-") + str(uuid4())
           
            imwrite(
                os.path.join(
                    real_mask_patch_dir,
                    name + eventid + str(count) + self.pattern,
                ),
                self.eroded_crop_labelimage.astype("uint16"),
            )

    def _region_selector(self):

        non_zero_indices = list(zip(*np.where(self.crop_labelimage > 0)))

        total_indices = list(zip(*np.where(self.crop_labelimage >= 0)))
        if len(total_indices) > 0:
            norm_foreground = len(non_zero_indices) / len(total_indices)
            index_ratio = float(norm_foreground)
            if (
                index_ratio >= self.lower_ratio_fore_to_back
                and index_ratio <= self.upper_ratio_fore_to_back
            ):

                self.valid = True


def erode_label_holes(lbl_img, iterations):
    lbl_img_filled = np.zeros_like(lbl_img)
    for lb in range(np.min(lbl_img), np.max(lbl_img) + 1):
        mask = lbl_img == lb
        mask_filled = binary_erosion(mask, iterations=iterations)
        lbl_img_filled[mask_filled] = lb
    return lbl_img_filled


def erode_labels(segmentation, erosion_iterations=2):
    # create empty list where the eroded masks can be saved to
    regions = regionprops(segmentation)
    erode = np.zeros(segmentation.shape)

    def erode_mask(segmentation_labels, label_id, erosion_iterations):

        only_current_label_id = np.where(segmentation_labels == label_id, 1, 0)
        eroded = binary_erosion(
            only_current_label_id, iterations=erosion_iterations
        )
        relabeled_eroded = np.where(eroded == 1, label_id, 0)
        return relabeled_eroded

    for i in range(len(regions)):
        label_id = regions[i].label
        erode = erode + erode_mask(segmentation, label_id, erosion_iterations)

    # convert list of numpy arrays to stacked numpy array
    return erode



import subprocess


def update_version_file():
    # Get the latest Git tag
    tag = (
        subprocess.check_output(["git", "describe", "--tags", "--abbrev=0"])
        .decode()
        .strip()
    )
    if "v" in tag:
        tag = tag.lstrip("v")
    # Update _version.py
    with open("src/vollseg/_version.py", "w") as version_file:
        version_file.write(f'__version__ = version = "{tag}"\n')
        version_file.write(
            f'__version_tuple__ = version_tuple = {tuple(map(int, tag.split(".")))}\n'
        )


if __name__ == "__main__":
    update_version_file()



#!/usr/bin/env python2
"""
Created on Mon Mar  8 16:22:00 2021

@author: vkapoor
"""
import numpy as np
from tqdm import tqdm
from csbdeep.utils import save_json
from csbdeep.utils import normalize
from stardist.matching import matching_dataset
from scipy.optimize import minimize_scalar
from csbdeep.utils import _raise
from .utils import VollSeg


class OptimizeThreshold:
    def __init__(
        self,
        X,
        Y,
        basedir,
        star_model,
        UseProbability=True,
        unet_model=None,
        noise_model=None,
        seedpool=True,
        dounet=True,
        n_tiles=(1, 1, 1),
        min_size=20,
        nms_threshs=[0, 0.3, 0.4, 0.5],
        iou_threshs=[0.3, 0.5, 0.7],
        measure="accuracy",
        RGB=False,
        axes="ZYX",
    ):

        self.star_model = star_model
        self.unet_model = unet_model
        self.noise_model = noise_model
        self.n_tiles = n_tiles
        self.basedir = basedir
        self.UseProbability = UseProbability
        self.X = X
        self.Y = Y
        self.RGB = RGB
        self.dounet = dounet
        self.seedpool = seedpool
        self.axes = axes
        self.nms_threshs = nms_threshs
        self.iou_threshs = iou_threshs
        self.measure = measure

        self.min_size = min_size

        self.Optimize()

    def Optimize(self):

        self.Y = [y for y in tqdm(self.Y)]
        self.X = [normalize(x, 1, 99.8, axis=(0, 1)) for x in tqdm(self.X)]

        print("Images to apply prediction on", len(self.X))

        (
            opt_prob_thresh_voll,
            opt_measure_voll,
            opt_nms_thresh,
            opt_prob_thresh_star,
            _,
        ) = (None, -np.inf, None, None, -np.inf)
        for _opt_nms_thresh in self.nms_threshs:
            (
                _opt_prob_thresh_voll,
                _opt_measure_voll,
                _opt_prob_thresh_star,
                _opt_measure_star,
            ) = self.optimize_threshold(self.Y, nms_thresh=_opt_nms_thresh)
            if _opt_measure_voll > opt_measure_voll:
                (
                    opt_prob_thresh_voll,
                    opt_measure_voll,
                    opt_nms_thresh,
                    opt_prob_thresh_star,
                    _,
                ) = (
                    _opt_prob_thresh_voll,
                    _opt_measure_voll,
                    _opt_nms_thresh,
                    _opt_prob_thresh_star,
                    _opt_measure_star,
                )
        opt_threshs_voll = dict(prob=opt_prob_thresh_voll, nms=opt_nms_thresh)
        opt_threshs_star = dict(prob=opt_prob_thresh_star, nms=opt_nms_thresh)

        self.thresholds_voll = opt_threshs_voll
        self.thresholds_star = opt_threshs_star
        print(
            "Using optimized values for vollseg: prob_thresh={prob:g}, nms_thresh={nms:g}.",
            opt_threshs_voll,
        )
        if self.basedir is not None:
            print("Saving to 'thresholds_voll.json'.")
            save_json(
                opt_threshs_voll, str(self.basedir + "/" + "thresholds_voll.json")
            )
        print(
            "Using optimized values for stardist: prob_thresh={prob:g}, nms_thresh={nms:g}.",
            opt_threshs_star,
        )
        if self.basedir is not None:
            print("Saving to 'thresholds_star.json'.")
            save_json(
                opt_threshs_star, str(self.basedir + "/" + "thresholds_star.json")
            )

        return opt_threshs_voll, opt_threshs_star

    def optimize_threshold(
        self, Y, nms_thresh, measure="accuracy", tol=1e-2, maxiter=20, verbose=1
    ):
        """Tune prob_thresh for provided (fixed) nms_thresh to maximize matching score (for given measure and averaged over iou_threshs)."""
        np.isscalar(nms_thresh) or _raise(ValueError("nms_thresh must be a scalar"))
        self.iou_threshs = (
            [self.iou_threshs] if np.isscalar(self.iou_threshs) else self.iou_threshs
        )
        values_voll = dict()
        values_star = dict()
        with tqdm(
            total=maxiter,
            disable=(verbose != 1),
            desc="NMS threshold = %g" % nms_thresh,
        ) as progress:

            def fn(thr):
                prob_thresh = thr
                value_voll = values_voll.get(prob_thresh)
                value_star = values_star.get(prob_thresh)
                if value_voll is None:
                    res = tuple(
                        zip(
                            *tuple(
                                VollSeg(
                                    x,
                                    unet_model=self.unet_model,
                                    star_model=self.star_model,
                                    axes=self.axes,
                                    noise_model=self.noise_model,
                                    n_tiles=self.n_tiles,
                                    UseProbability=self.UseProbability,
                                    dounet=self.dounet,
                                    seedpool=self.seedpool,
                                    RGB=self.RGB,
                                )
                                for x in tqdm(self.X)
                            )
                        )
                    )

                    if self.noise_model is None and self.star_model is not None:
                        (
                            Sizedsmart_seeds,
                            SizedMask,
                            star_labels,
                            proabability_map,
                            Markers,
                            Skeleton,
                        ) = res

                    elif self.noise_model is not None and self.star_model is not None:
                        (
                            Sizedsmart_seeds,
                            SizedMask,
                            star_labels,
                            proabability_map,
                            Markers,
                            Skeleton,
                            image,
                        ) = res

                    elif self.star_model is None:

                        raise ValueError(
                            f"StarDist model can not be {self.star_model} for evaluating optimized threshold"
                        )

                    stats_voll = matching_dataset(
                        Y,
                        Sizedsmart_seeds,
                        thresh=self.iou_threshs,
                        show_progress=False,
                        parallel=True,
                    )
                    values_voll[prob_thresh] = value_voll = np.mean(
                        [s._asdict()[measure] for s in stats_voll]
                    )

                    stats_star = matching_dataset(
                        Y,
                        star_labels,
                        thresh=self.iou_threshs,
                        show_progress=False,
                        parallel=True,
                    )
                    values_star[prob_thresh] = value_star = np.mean(
                        [s._asdict()[measure] for s in stats_star]
                    )

                    progress.update()
                    progress.set_postfix_str(
                        f"VollSeg-StarDist, {prob_thresh:.3f} -> {value_voll:.3f}, {value_star:.3f}"
                    )

                    progress.refresh()

                return -value_voll, -value_star

            opt_voll, opt_star = minimize_scalar(
                fn, method="golden", tol=tol, options={"maxiter": maxiter}
            )

        return opt_voll.x, -opt_voll.fun, opt_star.x, -opt_star.fun


def _is_floatarray(x):
    return isinstance(x.dtype.type(0), np.floating)
